@@ -1,22 +1,25 @@
-use std::collections::BTreeSet;
+use std::collections::HashSet;
 use std::str::FromStr;
 
 use proc_macro2::{Ident, TokenStream};
 use proc_macro_error::emit_error;
 use quote::{quote_spanned, ToTokens};
-use syn::{Lit, Abi, Attribute, Block, Error, Expr, FnArg, GenericParam, Generics, ImplItemMethod, Item, ItemImpl, ItemMod, ItemStruct, Lifetime, LifetimeDef, LitStr, parse_quote, Pat, PatIdent, PatType, ReturnType, Signature, Type, TypeReference, Visibility, VisPublic, Meta, NestedMeta};
+use syn::{Abi, Attribute, Block, Error, Expr, FnArg, GenericParam, Generics, ImplItemMethod, Item, ItemImpl, ItemMod, ItemStruct, Lifetime, LifetimeDef, Lit, LitStr, parse_quote, Pat, PatIdent, PatType, ReturnType, Signature, Type, TypeReference, Visibility, VisPublic, Meta, Path};
 use syn::fold::Fold;
-use syn::parse::{Parse, ParseStream};
+use syn::parse::{Parse, ParseStream, Parser};
 use syn::punctuated::Punctuated;
 use syn::spanned::Spanned;
 use syn::Token;
 use syn::token::Extern;
 use syn::visit::Visit;
 
+use darling::FromMeta;
+
 use proc_macro_error::{emit_warning};
 
 use crate::utils::unique_ident;
 use crate::validation::JNIBridgeModule;
+use darling::util::Flag;
 
 pub(crate) struct ModTransformer {
     module: JNIBridgeModule
@@ -56,7 +59,7 @@ impl ModTransformer {
     /// If the impl block is a standard impl block for a type, makes every child item (i.e. every fn) a freestanding one
     fn transform_item_impl(&mut self, node: ItemImpl) -> TokenStream {
         let transformed_item_impl = if let Type::Path(p) = &*node.self_ty {
-            let struct_name = p.path.segments.last().unwrap().ident.to_string();
+            let struct_name = p.path.get_ident().expect(&format!("expected identifier in item impl, got: {}", p.path.to_token_stream())).to_string();
             let struct_package = self.module.package_map[&struct_name].clone();
             let mut impl_transformer = ImplTransformer { struct_name, package: struct_package };
 
@@ -138,108 +141,68 @@ impl Fold for ModTransformer {
 }
 
 #[derive(Clone)]
+struct JavaPath(String);
+
+impl FromMeta for JavaPath {
+    fn from_value(value: &Lit) -> darling::Result<Self> {
+        use darling::Error;
+
+        if let Lit::Str(literal) = value {
+            let path = literal.value();
+            if path.contains('-') {
+                Err(Error::custom("invalid path: packages and classes cannot contain dashes"))
+            } else {
+                let tokens = TokenStream::from_str(&path).map_err(|_| Error::custom("cannot create token stream for java path parsing"))?;
+                let _parsed: Punctuated<Ident, Token![.]> = Punctuated::<Ident, Token![.]>::parse_separated_nonempty.parse(tokens.into()).map_err(|e| Error::custom(format!("cannot parse java path ({})", e)))?;
+
+                Ok(JavaPath(path.to_string()))
+            }
+        } else {
+            Err(Error::custom("invalid type"))
+        }
+    }
+}
+
+#[derive(Clone, Default, FromMeta)]
+#[darling(default)]
 struct SafeParams {
-    exception_class: Option<String>,
+    exception_class: Option<JavaPath>,
     message: Option<String>,
 }
 
-#[derive(Clone)]
+#[derive(Clone, FromMeta)]
 enum CallType {
     Safe(Option<SafeParams>),
-    Unchecked,
+    Unchecked(Flag),
 }
 
 impl Parse for CallType {
     fn parse(input: ParseStream<'_>) -> syn::Result<Self> {
-        let attribute = input.call(Attribute::parse_outer)?.first().cloned().expect("Invalid parsing of `call_type` attribute ");
+        let attribute = input.call(Attribute::parse_outer)?.first().cloned().ok_or_else(|| Error::new(input.span(), "Invalid parsing of `call_type` attribute "))?;
 
-        if attribute.path.get_ident().unwrap() != "call_type" {
-            panic!("expected identifier `call_type` for attribute")
+        if attribute.path.get_ident().ok_or_else(|| Error::new(attribute.path.span(), "expected identifier for attribute"))? != "call_type" {
+            return Err(Error::new(attribute.path.span(), "expected identifier `call_type` for attribute"));
         }
 
-        let args = attribute.parse_meta().expect("Meta parsing of `call_type` attribute tokens failed");
+        let attr_meta: Meta = attribute.parse_meta()?;
 
-        // TODO: Replace this manual, incomplete argument parsing monster with something sensible that can accept something like #[call_type(safe(exception_class = "java.io.IOException", message = "foobar"))]
-        match args {
-            Meta::List(args) => {
-                match args.path.get_ident().map(ToString::to_string).as_ref().map(String::as_str) {
-                    Some("call_type") => {
-                        args.nested.iter().next().map(|arg| {
-                            match arg {
-                                NestedMeta::Meta(arg) => {
-                                    match arg {
-                                        Meta::Path(arg) => {
-                                            match arg.get_ident().map(ToString::to_string).as_ref().map(String::as_str) {
-                                                Some("unchecked") => Ok(CallType::Unchecked),
-                                                Some("safe") => Ok(CallType::Safe(None)),
-                                                _ => Err(Error::new(arg.span(), format!("expected one of `safe`, `unchecked`, found {}", arg.into_token_stream())))
-                                            }
-                                        }
-
-                                        Meta::List(arg) => {
-                                            match arg.path.get_ident().map(ToString::to_string).as_ref().map(String::as_str) {
-                                                Some("safe") => {
-                                                    let mut exception_class = None;
-                                                    let mut message = None;
-
-                                                    arg.nested.iter().for_each(|arg| {
-                                                        match arg {
-                                                            NestedMeta::Meta(arg) => {
-                                                                match arg {
-                                                                    Meta::NameValue(arg) => {
-                                                                        match arg.path.get_ident().map(ToString::to_string).as_ref().map(String::as_str) {
-                                                                            Some("message") => {
-                                                                                match &arg.lit {
-                                                                                    Lit::Str(lit) => message = Some(lit.value()),
-                                                                                    _ => emit_error!(arg.lit, format!("expected string literal, found {}", arg.lit.to_token_stream()))
-                                                                                }
-                                                                            }
-
-                                                                            Some("exception_class") => {
-                                                                                match &arg.lit {
-                                                                                    // TODO: Proper parsing of java paths
-                                                                                    Lit::Str(lit) => exception_class = Some(lit.value().replace('.', "/")),
-                                                                                    _ => emit_error!(arg.lit, format!("expected string literal, found {}", arg.lit.to_token_stream()))
-                                                                                }
-                                                                            }
-
-                                                                            _ => emit_error!(arg, format!("invalid `safe` option: {}", arg.into_token_stream()))
-                                                                        }
-                                                                    }
-                                                                    _ => emit_error!(arg.span(), format!("expected name-value pair, found {}", arg.into_token_stream()))
-                                                                }
-                                                            }
-                                                            _ => emit_error!(arg.span(), format!("expected name-value pair, found nested {}", arg.into_token_stream()))
-                                                        }
-                                                    });
-
-                                                    Ok(CallType::Safe(Some(SafeParams { exception_class, message })))
-                                                }
-                                                _ => Err(Error::new(arg.span(), format!("expected `safe(..)`, found {}", arg.into_token_stream())))
-                                            }
-                                        }
-                                        _ => Err(Error::new(arg.span(), "invalid argument options supplied"))
-                                    }
-                                }
-                                _ => Err(Error::new(arg.span(), "invalid argument options format supplied"))
-                            }
-                        }).transpose().map(|i| i.unwrap())
-                    },
-                    _ => Err(Error::new(args.path.span(), "expected `call_type` identifier on attribute"))
-                }
-            }
-            _ => Err(Error::new(args.span(), "bug -- please report to library author. `call_type` parser supplied to invalid argument"))
+        // Special-case `call_type(safe)` without further parentheses
+        // TODO: Find out if it's possible to use darling to allow `call_type(safe)` *and* `call_type(safe(message = "foo"))` etc.
+        if attr_meta.to_token_stream().to_string() == "call_type(safe)" {
+            Ok(CallType::Safe(None))
+        } else {
+            CallType::from_meta(&attr_meta).map_err(|e| Error::new(attr_meta.span(), format!("invalid `call_type` attribute options ({})", e)))
         }
     }
 }
 
 struct AttributeFilter<'ast> {
-    pub whitelist: BTreeSet<String>,
+    pub whitelist: HashSet<Path>,
     pub filtered_attributes: Vec<&'ast Attribute>,
 }
 
 impl<'ast> AttributeFilter<'ast> {
-    fn with_whitelist(whitelist: BTreeSet<String>) -> Self {
+    fn with_whitelist(whitelist: HashSet<Path>) -> Self {
         AttributeFilter {
             whitelist,
             filtered_attributes: Vec::new(),
@@ -249,7 +212,7 @@ impl<'ast> AttributeFilter<'ast> {
 
 impl<'ast> Visit<'ast> for AttributeFilter<'ast> {
     fn visit_attribute(&mut self, attribute: &'ast Attribute) {
-        if self.whitelist.contains(&attribute.path.segments.first().unwrap().ident.to_string()) {
+        if self.whitelist.contains(&attribute.path) {
             self.filtered_attributes.push(attribute);
         }
     }
@@ -263,8 +226,8 @@ struct ImplTransformer {
 impl Fold for ImplTransformer {
     fn fold_impl_item_method(&mut self, node: ImplItemMethod) -> ImplItemMethod {
         let whitelist = {
-            let mut f = BTreeSet::new();
-            f.insert("call_type".into());
+            let mut f = HashSet::new();
+            f.insert(syn::parse2(TokenStream::from_str("call_type").unwrap()).unwrap());
             f
         };
 
@@ -273,7 +236,7 @@ impl Fold for ImplTransformer {
 
         let call_type_attribute = attributes_collector.filtered_attributes.first().and_then(|call_type_attr| {
             syn::parse2(call_type_attr.to_token_stream()).map_err(|e| {
-                emit_warning!(call_type_attr, format!("invalid parsing of `call_type` attribute, defaulting to #[call_type(safe)]. Error in lifted args parsing: {}", e));
+                emit_warning!(e.span(), format!("invalid parsing of `call_type` attribute, defaulting to #[call_type(safe)]. {}", e));
                 e
             }).ok()
         }).unwrap_or(CallType::Safe(None));
@@ -336,7 +299,7 @@ impl Fold for ImplMethodTransformer {
                 let input_param: Expr = {
                     match self.call_type {
                         CallType::Safe(_) => parse_quote! { TryFromJavaValue::try_from(#arg, &env)? },
-                        CallType::Unchecked => parse_quote! { FromJavaValue::from(#arg, &env) }
+                        CallType::Unchecked { .. } => parse_quote! { FromJavaValue::from(#arg, &env) }
                     }
                 };
                 input_param
@@ -378,7 +341,7 @@ impl Fold for ImplMethodTransformer {
             let outer_signature_span = s.span();
             let outer_output_type: Type = match s.output {
                 ReturnType::Default => parse_quote!(()),
-                ReturnType::Type(_, ty) => *ty.clone()
+                ReturnType::Type(_, ty) => *ty
             };
 
             s.output = ReturnType::Type(Token![->](outer_signature_span), Box::new(parse_quote!(::jni::errors::Result<#outer_output_type>)));
@@ -389,7 +352,7 @@ impl Fold for ImplMethodTransformer {
         let method_impl = node.block;
 
         let new_block: Block = match &self.call_type {
-            CallType::Unchecked => parse_quote! {{
+            CallType::Unchecked { .. } => parse_quote! {{
                 #inner_signature
                     #method_impl
 
@@ -397,10 +360,15 @@ impl Fold for ImplMethodTransformer {
             }},
 
             CallType::Safe(exception_details) => {
-                let default_params = SafeParams { exception_class: Some("java/lang/RuntimeException".into()), message: Some("JNI conversion error!".into()) };
-                let (exception_class, message) = {
-                    let SafeParams { exception_class: e, message: m } = exception_details.clone().unwrap_or(default_params.clone());
-                    (e.unwrap_or(default_params.exception_class.unwrap()), m.unwrap_or(default_params.message.unwrap()))
+                let (default_exception_class, default_message) = ("java/lang/RuntimeException", "JNI conversion error!");
+                let (exception_class, message) = match exception_details {
+                    Some(SafeParams { exception_class, message }) => {
+                        let exception_class_result = exception_class.as_ref().map(|v| &v.0).map(AsRef::as_ref).unwrap_or(default_exception_class);
+                        let message_result = message.as_ref().map(AsRef::as_ref).unwrap_or(default_message);
+
+                        (exception_class_result, message_result)
+                    }
+                    None => (default_exception_class, default_message)
                 };
 
                 parse_quote! {{
@@ -456,7 +424,7 @@ impl Fold for ImplMethodTransformer {
                 let mut s = s.replace('.', "_");
                 s.push('_');
                 s
-            }).unwrap_or("".into());
+            }).unwrap_or_else(|| "".into());
 
             format!("Java_{}{}_{}", snake_case_package, self.struct_name, node.ident.to_string())
         };
@@ -532,7 +500,7 @@ impl Fold for FreestandingTransformer {
                 };
 
                 FnArg::Typed(PatType {
-                    attrs: r.attrs.clone(),
+                    attrs: r.attrs,
                     pat: Box::new(Pat::Ident(PatIdent {
                         attrs: vec![],
                         by_ref: None,
@@ -598,7 +566,7 @@ impl Fold for JNISignatureTransformer {
 
                 let jni_conversion_type: Type = match self.call_type {
                     CallType::Safe(_) => syn::parse2(quote_spanned! { original_input_type.span() => <#original_input_type as TryFromJavaValue<'env>>::Source }).unwrap(),
-                    CallType::Unchecked => syn::parse2(quote_spanned! { original_input_type.span() => <#original_input_type as FromJavaValue<'env>>::Source }).unwrap(),
+                    CallType::Unchecked { .. } => syn::parse2(quote_spanned! { original_input_type.span() => <#original_input_type as FromJavaValue<'env>>::Source }).unwrap(),
                 };
 
                 FnArg::Typed(PatType {
@@ -630,7 +598,7 @@ impl Fold for JNISignatureTransformer {
             ReturnType::Default => return_type,
             ReturnType::Type(ref arrow, ref rtype) => {
                 match (&**rtype, self.call_type.clone()) {
-                    (Type::Path(p), CallType::Unchecked) => {
+                    (Type::Path(p), CallType::Unchecked { .. }) => {
                         ReturnType::Type(*arrow, syn::parse2(quote_spanned! { p.span() => <#p as IntoJavaValue<'env>>::Target }).unwrap())
                     }
 
@@ -638,7 +606,7 @@ impl Fold for JNISignatureTransformer {
                         ReturnType::Type(*arrow, syn::parse2(quote_spanned! { p.span() => <#p as TryIntoJavaValue<'env>>::Target }).unwrap())
                     }
 
-                    (Type::Reference(r), CallType::Unchecked) => {
+                    (Type::Reference(r), CallType::Unchecked { .. }) => {
                         ReturnType::Type(*arrow, syn::parse2(quote_spanned! { r.span() => <#r as IntoJavaValue<'env>>::Target }).unwrap())
                     }
 
